/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;
import javassist.expr.MethodCall;
import javassist.expr.FieldAccess;
import javassist.expr.NewExpr;
import javassist.gluonj.pc.*;
import javassist.gluonj.weave.Advice;
import javassist.gluonj.weave.GluonMaker;
import javassist.gluonj.weave.Inspector;
import javassist.gluonj.weave.Refiner;
import javassist.gluonj.weave.AliasList;
import javassist.gluonj.weave.Hook;
import javassist.gluonj.weave.Matcher;
import javassist.gluonj.weave.CallMatcher;
import javassist.gluonj.weave.FieldMatcher;
import javassist.gluonj.weave.NewMatcher;
import javassist.gluonj.weave.CflowCollector;

/**
 * A module of pointcuts and advices. 
 */
public class Gluon {
    private String glueName;
    private ClassPool pool;
    protected List children;        // List<Gluon>
    protected List refiners;          // List<Refiner>
    private HashMap helpers;        // HashMap<String helperClassName,String>
    protected List advices;           // List<Advice>
    private ArrayList inspectors;   // ArrayList<Inspector>

    private static final int UNKNOWN = 0;
    private static final int EMPTY = -1;
    private static final int NOT_EMPTY = 1;
    private int hasAdviceCache;

    /**
     * A factory object of <code>Gluon</code> object. 
     */
    public static GluonMaker maker = new GluonMaker();

    /**
     * True if a stack trace should be printed when an error message
     * is shown.
     */
    public static boolean stackTrace = false;

    /**
     * A special form for converting a reference type.
     */
    public static Object $refine(Object ref) {
        return ref;
    }

    /**
     * Displays version information and exits.
     */
    public static void main(String[] args) {
        if (args.length == 0) {
            System.out.println("GluonJ version 1.4");
            System.out.println("Copyright (C) 2006-2007 by Shigeru Chiba.  All Rights Reserved.");
            CtClass.main(args);
        }
        else
            javassist.gluonj.weave.OfflineWeaver.run("-jar gluonj.jar", args);
    }

    /**
     * Constructs a module.
     */
    public Gluon(String name) {
        this(name, new ClassPool(true));
    }

    /**
     * Constructs a module with a given <code>ClassPool</code>.
     */
    public Gluon(String name, ClassPool cp) {
        pool = cp;
        glueName = name;
        children = new LinkedList();
        refiners = new LinkedList();
        helpers = new HashMap();
        advices = new LinkedList();
        hasAdviceCache = UNKNOWN;
    }

    /**
     * Returns a <code>ClassPool</code> that was given to a constructor.
     */
    public ClassPool getPool() { return pool; }

    /**
     * Obtains the name of the glue.
     */
    public String getName() { return glueName; }

    /**
     * Returns the string representation of the glue.
     */
    public String toString() {
        return super.toString() + "[" + glueName + "]"; 
    }

    /**
     * Adds a refine class.
     * 
     * @param refine    refine class.
     */
    public Gluon refine(CtClass refine) {
        // this does not try to find the target class.
        // It may not exist.
        refiners.add(new Refiner(refine, glueName, false, null));
        return this;
    }

    /**
     * Adds a refine class.
     * 
     * @param refine    refine class.
     * @param anno      annotations or null.
     */
    public Gluon refine(CtClass refine, Object[] anno) {
        if (anno == null)
            anno = refine.getAvailableAnnotations();

        boolean priv = false;
        String target = null;
        for (int i = 0; i < anno.length; i++)
            if (anno[i] instanceof Privileged)
                priv = true;
            else if (anno[i] instanceof Refine) {
                String arg = ((Refine)anno[i]).value();
                if (arg != null && arg.length() > 0)
                    target = arg;
            }

        refiners.add(new Refiner(refine, glueName, priv, target));
        return this;
    }

    /**
     * Adds a helper class.
     */
    public void addHelper(String className) {
        helpers.put(className, className);
    }

    /**
     * Returns true if the class specified by <code>className</code>
     * is a hleper class.
     */
    public boolean isHelper(String className) {
        return helpers.get(className) != null;
    }

    /**
     * Returns the names of the registered refine classes.
     *
     * @return a list of <code>Refiner</code>.  Its members are
     * a target class name, a refine class, and a glue name declaring. 
     */
    public Iterator getRefiners() {
        return new Iterator() {
            Iterator glues = children.iterator();
            boolean finish = false;
            Iterator cur = nextIterator();

            private Iterator nextIterator() {
                if (glues.hasNext()) {
                    Gluon g = (Gluon)glues.next();
                    return g.getRefiners();
                }
                else if (finish)
                    return null;
                else {
                    finish = true;
                    return refiners.iterator();
                }
            }

            public boolean hasNext() {
                do {
                    if (cur.hasNext())
                        return true;

                    cur = nextIterator();
                } while (cur != null);
                return false;
            }

            public Object next() {
                return cur.next();
            }

            public void remove() {
                throw new RuntimeException("fatal: not implemented");
            }
        };
    }

    /**
     * Returns true if this glue (or its child) has an advice.
     */
    public boolean hasAdvice() {
        if (hasAdviceCache != UNKNOWN)
            return hasAdviceCache == EMPTY ? false : true;

        if (!advices.isEmpty()) {
            hasAdviceCache = NOT_EMPTY;
            return true;
        }

        Iterator includes = children.iterator();
        while (includes.hasNext()) {
            Gluon g = (Gluon)includes.next();
            if (g.hasAdvice()) {
                hasAdviceCache = NOT_EMPTY;
                return true;
            }
        }

        hasAdviceCache = EMPTY;
        return false;
    }

    /**
     * Adds a child glue.  The glue earlier registered has a higher priority. 
     */
    public Gluon include(Gluon child) {
        children.add(child);
        hasAdviceCache = UNKNOWN;
        helpers.putAll(child.helpers);
        return this;
    }

    public boolean isIncluded(String glueName) {
        if (glueName.equals(getName()))
            return true;

        Iterator it = children.iterator();
        while (it.hasNext()) {
            Gluon c = (Gluon)it.next();
            if (c.isIncluded(glueName))
                return true; 
        }

        return false;
    }

    /**
     * Makes a new advice and apends it to the glue.
     *
     * @param advice        the body of the advice.
     * @return the appended advice.
     */
    public Advice addAdvice(int kind, Pointcut pc, String advice)
        throws WeaveException
    {
        AliasList alias = new AliasList();
        PointcutNode pcn = pc.parse(this, alias);
        Advice adv = new Advice(kind, pcn, alias.next, advice);
        advices.add(adv);
        hasAdviceCache = UNKNOWN;
        return adv;
    }

    private void addAdvice2(int kind, Pointcut pc, String advice) {
        try {
            addAdvice(kind, pc, advice);
        }
        catch (WeaveException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    /**
     * Adds a before advice.  It throws a <code>RuntimeException</code>
     * if a problem occurs.
     */
    public Gluon before(Pointcut pc, String advice) {
        addAdvice2(Advice.BEFORE, pc, advice);
        return this;
    }

    /**
     * Adds a before advice.
     */
    public Gluon addBefore(Pointcut pc, String advice)
        throws WeaveException
    {
        addAdvice(Advice.BEFORE, pc, advice);
        return this;
    }

    /**
     * Adds an after advice.  It throws a <code>RuntimeException</code>
     * if a problem occurs.
     */
    public Gluon after(Pointcut pc, String advice) {
        addAdvice2(Advice.AFTER, pc, advice);
        return this;
    }

    /**
     * Adds an after advice.
     */
    public Gluon addAfter(Pointcut pc, String advice)
        throws WeaveException
    {
        addAdvice(Advice.AFTER, pc, advice);
        return this;
    }

    /**
     * Adds an around advice.  It throws a <code>RuntimeException</code>
     * if a problem occurs.
     */
    public Gluon around(Pointcut pc, String advice) {
        addAdvice2(Advice.AROUND, pc, advice);
        return this;
    }

    /**
     * Adds an around advice.
     */
    public Gluon addAround(Pointcut pc, String advice)
        throws WeaveException
    {
        addAdvice(Advice.AROUND, pc, advice);
        return this;
    }

    /**
     * Creates a glue according to a given annotated class.
     * All exceptions thrown during the creation are converted into
     * a RuntimeException.
     *
     * @param className     the name of the annotated class.
     */
    public static Gluon create(String className) throws RuntimeException {
        try {
            return make(className);
        }
        catch (WeaveException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    /**
     * Makes a glue according to a given annotated class.
     *
     * @param className     the name of the annotated class.
     */
    public static Gluon make(String className) throws WeaveException {
        return make(className, new ClassPool(true), null);
    }

    /**
     * Makes a glue according to a given annotated class.
     *
     * @param className     the name of the annotated class.
     * @param cp            the class pool used for obtaining the
     *                      definition of the annotated class.
     */
    public static Gluon make(String className, ClassPool cp)
        throws WeaveException
    {
        return make(className, cp, null);
    }

    /**
     * Makes a glue according to a given annotated class.
     *
     * @param className     the name of the annotated class.
     * @param cp            the class pool used for obtaining the
     *                      definition of the annotated class.
     * @param cloader       the class loader used for making an instance of the
     *                      annotated class.  It can be null.
     */
    public static Gluon make(String className, ClassPool cp, ClassLoader cloader)
        throws WeaveException
    {
        try {
            return maker.make(className, cp, cloader, null);
        }
        catch (NotFoundException e) {
            throw new WeaveException(e);
        }
    }

    /**
     * Appends an inspector to the list of registered inspectors.
     * The inspectors are called for every class definition before
     * the class is woven.
     */
    public void registerInspector(Inspector i) {
        if (inspectors == null)
            inspectors = new ArrayList();

        inspectors.add(i);
    }

    /**
     * Returns a list of inspectors.
     *
     * @return  null    if there is no inspector.
     */
    public List getInspectors() { return inspectors; }

    /**
     * Returns advices that should be executed at a given join point.
     */
    public Hook.List match(MethodCall mcall) throws WeaveException {
        return match(new CallMatcher(mcall));
    }

    /**
     * Returns advices that should be executed at a given join point.
     */
    public Hook.List match(FieldAccess fa) throws WeaveException {
        return match(new FieldMatcher(fa));
    }

    /**
     * Returns advices that should be executed at a given join point.
     */
    public Hook.List match(NewExpr newexpr) throws WeaveException {
        return match(new NewMatcher(newexpr));
    }

    private Hook.List match(Matcher matcher) throws WeaveException {
        Hook.List list = new Hook.List();
        match(matcher, list);
        return list;
    }

    private void match(Matcher matcher, Hook.List list) throws WeaveException {
        Iterator includes = children.iterator();
        while (includes.hasNext()) {
            Gluon g = (Gluon)includes.next();
            g.match(matcher, list);
        }

        Iterator it = advices.iterator();
        while (it.hasNext()) {
            Advice adv = (Advice)it.next();
            adv.getPointcut().accept(matcher);
            if (matcher.getResult())
                list.add(matcher.getResidue(), adv);
        }
    }

    /**
     * Returns all cflow() included in this glue.
     */
    public List findCflows() throws WeaveException {
        List list = new LinkedList();
        findCflows(list);
        return list;
    }

    private void findCflows(List list) throws WeaveException {
        CflowCollector collector = new CflowCollector(list);
        Iterator includes = children.iterator();
        while (includes.hasNext()) {
            Gluon g = (Gluon)includes.next();
            g.findCflows(list);
        }

        Iterator it = advices.iterator();
        while (it.hasNext()) {
            Advice adv = (Advice)it.next();
            adv.getPointcut().accept(collector);
        }
    }
}
