/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj;

import javassist.gluonj.pc.*;
import javassist.gluonj.weave.AliasList;
import java.io.Serializable;

public class Pointcut implements Serializable {
    private PointcutNode tree;
    private Pointcut prev;
    protected AliasList aliases;
    public Pointcut not;    // NOT is stronger than AND
    public Pointcut and;    // AND is stronger than OR
    public Pointcut or;

    protected Pointcut(Pointcut back) {
        this(back, false);
    }

    protected Pointcut(Pointcut back, boolean withNot) {
        tree = null;
        prev = back;
        aliases = null;
        or = null;
        and = null;
        if (withNot)
            not = new Pointcut(this, false);
        else
            not = null;
    }

    protected boolean isNot() { return false; }

    public String toString() {
        AliasList a = new AliasList();
        String str = parse(a).toString();
        if (a.next == null)
            return str;
        else
            return a.next.toString() + " " + str; 
    }

    static class Iterator {
        Pointcut node;
    }

    protected PointcutNode parse(Gluon gluon, AliasList a)
        throws WeaveException
    {
        PointcutNode pcn = parse(a);
        prepare(new PatternParser(gluon), pcn);
        return pcn;
    }

    private static void prepare(PatternParser parser, PointcutNode pcn)
        throws WeaveException
    {
        pcn.accept(new PrepareVisitor(parser));
    }

    protected PointcutNode parse(AliasList allAliases) {
        Iterator it = new Iterator();
        it.node = this;
        OrPc expr = parseOr(it, null, allAliases);
        if (expr.rest() == null)
            return expr.first();
        else
            return expr;
    }

    OrPc parseOr(Iterator it, OrPc tail, AliasList allAliases) {
        Pointcut pc;
        do {
            AndPc left = parseAnd(it, null, allAliases);
            tail = new OrPc(left.rest() == null ? left.first() : left, tail);
            if ((pc = it.node) == null)
                break;

            it.node = pc.prev;
        } while (it.node != null && it.node.or == pc);
        return tail;
    }

    AndPc parseAnd(Iterator it, AndPc tail, AliasList allAliases) {
        Pointcut pc;
        do {
            if ((pc = it.node) == null)
                break;

            boolean isNotPc = pc.prev == null ? pc.isNot() : pc.prev.not == pc;
            PointcutNode tree;
            if (isNotPc)
                tree = new NotPc(pc.tree);
            else
                tree = pc.tree;

            tail = new AndPc(tree, tail);
            if (pc.aliases != null)
                allAliases.next = AliasList.concat(allAliases.next, pc.aliases);

            if (isNotPc && pc.prev != null)
                pc = pc.prev;

            it.node = pc.prev;
        } while (it.node != null && it.node.and == pc);
        it.node = pc;
        return tail;
    }

    protected void setTree(PointcutNode node) {
        if (tree != null)
            throw new RuntimeException("syntax error in pointcut");

        tree = node;
        and = new Pointcut(this, true);
        or = new Pointcut(this, true);
        not = null;
    }

    Pointcut setAlias0(AliasList a) {
        aliases = a;
        return this;
    }

    /* define() is available only before other pointcut designators
     * are called.
     */
    public Pointcut define(String aliasName, String expr) {
        throw new RuntimeException("syntax error in define");
    }

    /* called only from Pcd.define() and AliasedPcExpr.define().
     */
    void define0(String aliasName, String expr) {
        AliasList a = new AliasList(aliasName, expr);
        aliases = AliasList.concat(aliases, a);
    }

    public Pointcut expr(Pointcut pc) {
        AliasList a = new AliasList();
        setTree(pc.parse(a));
        aliases = a.next;
        return this;
    }

    public Pointcut within(String className) {
        setTree(new WithinPc(className));
        return this;
    }

    public Pointcut call(String method) {
        setTree(new CallPc(method));
        return this;
    }

    public Pointcut set(String field) {
        setTree(new SetPc(field));
        return this;
    }

    public Pointcut get(String field) {
        setTree(new GetPc(field));
        return this;
    }

    /**
     * AspectJ's if pointcut designator.
     */
    public Pointcut when(String expr) {
        setTree(new IfPc(expr));
        return this;
    }

    public Pointcut cflow(String method) {
        setTree(new CflowPc(method));
        return this;
    }
}
