/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.plugin;

import javassist.NotFoundException;
import javassist.gluonj.*;
import javassist.gluonj.pc.*;
import javassist.gluonj.weave.*;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.annotation.Annotation;

/**
 * Implementation of annotate pointcut.
 */
@Glue public class MetaTag {

    /* Comment out the next line if you don't want to see a stack trace.
     */
    // @Refine
    static class Debug extends Gluon {
        public Debug() { super(null); }

        // print a stack trace when an error message is printed.
        public static boolean stackTrace = true;
    }

    public static class AnnotatePc extends PointcutNode {
        private String arg;
        private PcPattern pattern;

        public AnnotatePc(String pat) {
            this.arg = pat;
        }

        public String toString() {
            return "annotate(" + arg + ")";
        }

        public void prepare(PatternParser p) throws WeaveException {
            pattern = p.parseClass(removeAt(arg));
            if (!pattern.isClassName())
                throw new WeaveException("bad pattern: " + toString());
        }

        private static String removeAt(String name) {
            if (name != null && name.length() > 1 && name.charAt(0) == '@')
                return name.substring(1);
            else
                return name;
        }

        public void accept(PointcutVisitor v) throws WeaveException {
            ((PcVisitor2)v).visit(this);
        }

        public boolean match(AnnotationsAttribute attr) throws WeaveException {
            if (attr == null)
                return false;

            Annotation[] anno = attr.getAnnotations();
            int n = anno.length;
            for (int i = 0; i < n; i++)
                if (pattern.matchClass(anno[i].getTypeName()))
                    return true;

            return false;
        }
    }

    @Refine interface PcVisitor2 extends PointcutVisitor {
        void visit(AnnotatePc pc) throws WeaveException;
    } 

    @Refine static class Prepare2 extends PrepareVisitor {
        public Prepare2() { super(null); }
        public void visit(AnnotatePc pc) throws WeaveException {
            pc.prepare(parser);
        }
    }

    @Refine static abstract class Match2 extends Matcher {
        public void visit(AnnotatePc pc) throws WeaveException {
            result = false;
        }
    }

    @Refine static class CallMatch2 extends CallMatcher { 
        public CallMatch2() { super(null); }

        public void visit(AnnotatePc pc) throws WeaveException {
            try {
                MethodInfo minfo = joinPoint.getMethod().getMethodInfo2();
                AnnotationsAttribute aa1 = (AnnotationsAttribute)
                    minfo.getAttribute(AnnotationsAttribute.invisibleTag);
                AnnotationsAttribute aa2 = (AnnotationsAttribute)
                    minfo.getAttribute(AnnotationsAttribute.visibleTag);
                result = pc.match(aa1) || pc.match(aa2);
            }
            catch (NotFoundException e) {
                throw new WeaveException(e);
            }
        }
    }

    @Refine static class FieldMatch2 extends FieldMatcher {
        public FieldMatch2() { super(null); }

        public void visit(AnnotatePc pc) throws WeaveException {
            try {
                FieldInfo finfo = joinPoint.getField().getFieldInfo2();
                AnnotationsAttribute aa1 = (AnnotationsAttribute)
                    finfo.getAttribute(AnnotationsAttribute.invisibleTag);
                AnnotationsAttribute aa2 = (AnnotationsAttribute)
                    finfo.getAttribute(AnnotationsAttribute.visibleTag);
                result = pc.match(aa1) || pc.match(aa2);
            }
            catch (NotFoundException e) {
                throw new WeaveException(e);
            }
        }
    }

    @Refine static class Cflow2 extends CflowCollector {
        public Cflow2() { super(null); }

        public void visit(AnnotatePc pc) throws WeaveException {
            // do nothing.
        }
    }

    @Refine static class Pcut2 extends Pointcut {
        Pcut2() { super(null); }
        public Pointcut annotate(String tag) {
            setTree(new AnnotatePc(tag));
            return this;
        }
    }

    @Refine static class Pcd2 extends Pcd {
        public static Pointcut annotate(String tag) {
            Pcut2 pc = (Pcut2)make();
            return pc.annotate(tag);
        }
    }
}
