/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */

package javassist.gluonj.plugin;

import java.util.HashMap;
import javassist.*;
import javassist.expr.ExprEditor;
import javassist.expr.FieldAccess;
import javassist.expr.MethodCall;
import javassist.gluonj.*;
import javassist.gluonj.pc.PatternParser;
import javassist.gluonj.util.SimplePcNode;
import javassist.gluonj.weave.Residue;

/*
 * @author Muga Nishizawa
 * @author Shigeru Chiba
 */
@Glue public class Updater {
    @Refine static class Pcd2 extends Pcd {
        public static Pointcut updater(String className, String methodName) {
            Pointcut pc = make();
            return ((Pcut2)pc).updater(className, methodName);
        }
    }

    @Refine static class Pcut2 extends Pointcut {
        public Pcut2() { super(null); }
        public Pointcut updater(String className, String methodName) {
            setTree(new UpdaterPc(className, methodName));
            return this;
        }
    }

    public static class UpdaterPc extends SimplePcNode {
        private ClassPool cpool;
        private String rootClassName;
        private CtClass rootClass;
        private String methodName;
        private HashMap updatersMap;        // HashMap<String,HashMap<CtMethod,CtClass>>

        public UpdaterPc(String cName, String mName) {
            rootClassName = cName;
            methodName = mName;
            updatersMap = new HashMap();
        }

        public String toString() {
            return "updater(" + rootClassName + ", " + methodName + ")";
        }

        /**
         * Is invoked before GluonJ starts advice weaving.
         */
        public void prepare(PatternParser p) throws WeaveException {
            cpool = p.getClassPool();
            try {
                rootClass = cpool.get(rootClassName);
            }
            catch (NotFoundException e) {
                throw new WeaveException(e);
            }
        }

        /**
         * Collects updater methods declared in the given class.
         */
        public void inspect(CtClass clazz) throws WeaveException {
            if (!clazz.subclassOf(rootClass))
                return;

            try {
                String className = clazz.getName();
                CtMethod m = clazz.getDeclaredMethod(methodName);
                HashMap fields = accessedFields(className, m);
                HashMap updaters = updateMethods(clazz, fields);
                updatersMap.put(className, updaters);
            }
            catch (CannotCompileException e) {
                throw new WeaveException(e);
            }
            catch (NotFoundException e) {
                return;
            }
        }

        /**
         * Returns true if the given join point is a call to an updater method.
         */
        public boolean match(MethodCall joinPoint, Residue[] residue)
            throws WeaveException
        {
            String className = joinPoint.getClassName();
            Object updaters = updatersMap.get(className);
            if (updaters == null)
                return false;

            try {
                return ((HashMap)updaters).get(joinPoint.getMethod()) != null;
            }
            catch (NotFoundException e) {
                throw new WeaveException(e);
            }
        }

        /**
         * Returns the fields that the given method accesses during its execution. 
         */
        private HashMap accessedFields(final String className, CtMethod cm)
            throws CannotCompileException
        {
            final HashMap fields = new HashMap();
            cm.instrument(new ExprEditor() {
                public void edit(FieldAccess fa) throws CannotCompileException {
                    if (fa.getClassName().equals(className))
                        fields.put(fa.getFieldName(), className);
                }
            });
            return fields;
        }

        /**
         * Inspects the methods declared in the given class and
         * returns the methods that will update some of the given fields.
         */
        private HashMap updateMethods(CtClass clazz, HashMap fields)
            throws CannotCompileException
        {
            HashMap updaters = new HashMap();
            CtMethod[] methods = clazz.getDeclaredMethods();
            for (int i = 0; i < methods.length; i++) {
                MethodInspector mi = new MethodInspector(fields);
                methods[i].instrument(mi);
                if (mi.found)
                    updaters.put(methods[i], clazz);
            }

            return updaters;
        }

        /**
         * A nested class used by updateMethods().
         */
        private static class MethodInspector extends ExprEditor {
            private HashMap fields;
            public boolean found; 

            MethodInspector(HashMap map) {
                fields = map;
                found = false;
            }

            public void edit(FieldAccess fa) throws CannotCompileException {
                if (fa.isWriter()) {
                    String fname = fa.getFieldName();
                    Object className = fields.get(fname);
                    if (className != null && fa.getClassName().equals(className))
                        found = true;
                }
            }
        }
    }
}
