/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import javassist.*;
import javassist.expr.Expr;
import javassist.expr.ExprEditor;
import javassist.expr.MethodCall;
import javassist.expr.NewExpr;
import javassist.expr.FieldAccess;
import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;
import javassist.gluonj.pc.CflowPc;

import java.util.Iterator;
import java.util.List;

class WrappedWeaveException extends CannotCompileException {
    WeaveException weave;
    public WrappedWeaveException(WeaveException e) {
        super(e);
        weave = e;
    }
}

/**
 * Pointcut-advice processor.
 */
public class AdviceWeaver {
    public static final String CONSTRUCTOR = "<init>";

    public AdviceWeaver() {}

    public boolean weave(CtClass orig, final Gluon glue) throws WeaveException {
        if (!glue.hasAdvice())
            return false;

        WeaveEditor editor = new WeaveEditor(glue);
        List cflows = glue.findCflows();
        try {
            CtBehavior[] members = orig.getDeclaredBehaviors();
            for (int i = 0; i < members.length; i++) {
                members[i].instrument(editor);
                prepareCflow(cflows, members[i]);
            }
        }
        catch (WrappedWeaveException e) {
            throw e.weave;
        }
        catch (CannotCompileException e) {
            throw new WeaveException(e);
        }

        return editor.modified;
    }

    private void prepareCflow(List cflows, CtBehavior method)
        throws WeaveException
    {
        Iterator it = cflows.iterator();
        while (it.hasNext()) {
            CflowPc pc = (CflowPc)it.next();
            pc.instrument(method);
        }
    }

    /* To add a new kind of joinpoint, first declare edit() in WeaveEditor.
     * Then edit makeEditor().  Also, match() must be added to the Gluon class. 
     */
    public static class WeaveEditor extends ExprEditor {
        Gluon glue;
        boolean modified;

        public WeaveEditor(Gluon g) {
            glue = g;
            modified = false;
        }

        public void edit(MethodCall call) throws CannotCompileException {
            try {
                applyAdvices(glue.match(call), call);
            }
            catch (WeaveException e) {
                throw new WrappedWeaveException(e);
            }
        }

        public void edit(FieldAccess access) throws CannotCompileException {
            try {
                applyAdvices(glue.match(access), access);
            }
            catch (WeaveException e) {
                throw new WrappedWeaveException(e);
            }
        }

        public void edit(NewExpr access) throws CannotCompileException {
            try {
                applyAdvices(glue.match(access), access);
            }
            catch (WeaveException e) {
                throw new WrappedWeaveException(e);
            }
        }

        private void applyAdvices(Hook.List hooks, Expr expr)
            throws CannotCompileException
        {
            if (hooks.notEmpty()) {
                modified = true;
                insertHook(hooks.iterator(), expr);
            }
        }
    }

    public static void insertHook(Hook.Iterator iterator, Expr expr)
        throws CannotCompileException
    {
        StringBuffer before = new StringBuffer();
        StringBuffer after = new StringBuffer();
        while (iterator.hasNext()) {
            Hook hook = (Hook)iterator.next();
            if (insertHook(hook, iterator, expr, before, after))
                return;
        }

        before.insert(0, '{').append("$_ = $proceed($$);").append(after)
              .append('}');
        expr.replace(before.toString());
    }

    /**
     * Makes a hook code specified by <code>hook</code>.
     *
     * @return true if all other advices have been processed. 
     */
    protected static boolean insertHook(Hook hook, Hook.Iterator iterator,
                                     Expr expr, StringBuffer before, StringBuffer after)
        throws CannotCompileException
    {
        Advice adv = hook.advice;
        String body = adv.getBody();
        body = expandSpecialVars(body, expr);
        if (hook.residue != null) {
            String cond = hook.residue.getIf();
            if (cond != null) {
                StringBuffer sbuf = new StringBuffer();
                sbuf.append("if(").append(cond).append("){")
                    .append(body).append("}");
                if (adv.kind == Advice.AROUND)
                    sbuf.append("else{ $_ = $proceed($$); }");

                body = sbuf.toString();
            }
        }

        if (adv.kind == Advice.BEFORE)
            before.append('{').append(body).append('}');
        else if (adv.kind == Advice.AFTER)
            after.insert(0, '}').insert(0, body).insert(0, '{');
        else if (adv.kind == Advice.AROUND) {
            StringBuffer around = new StringBuffer(body);
            if (before.length() > 0)
                around.insert(0, before);

            if (after.length() > 0)
                around.append(after);

            ExprEditor ed;
            if (iterator.hasNext())
                ed = makeEditor(iterator, expr);
            else
                ed = null;

            if (Gluon.stackTrace)
                Weaver.logging("weave at " + expr.getLineNumber() + ':' + expr.getFileName() 
                               + " " + around.toString());

            expr.replace(around.toString(), ed);
            return true;
        }
        else
            throw new RuntimeException("bad advice: " + adv);

        return false;
    }

    private static String expandSpecialVars(String body, Expr expr)
        throws CannotCompileException
    {
        int len = body.length();
        int state = 0;      // 1: part of an identifier, 2: of a string literal.
        int start = 0;
        StringBuffer sbuf = new StringBuffer();
        for (int i = 0; i < len; i++) {
            char c = body.charAt(i);
            if (state == 0)
                if (Character.isJavaIdentifierStart(c)) {
                    state = 1;
                    start = i;
                }
                else {
                    sbuf.append(c);
                    if (c == '"')
                        state = 2;
                }
            else if (state == 1) {
                if (!Character.isJavaIdentifierPart(c)) {
                    String rep = null;
                    if (body.charAt(start) == '$')
                        rep = expandSpecialVariable(body.substring(start, i), expr);

                    if (rep == null)
                        for (int k = start; k < i; k++)
                            sbuf.append(body.charAt(k));
                    else
                        sbuf.append(rep);

                    i--;
                    state = 0;
                }
            }
            else {
                sbuf.append(c);
                if (c == '"')
                    state = 0;
                if (c == '\\') {
                    if (++i < len)
                        sbuf.append(body.charAt(i));
                }
            }
        }

        return sbuf.toString();
    }

    protected static String expandSpecialVariable(String varName, Expr expr)
        throws CannotCompileException
    {
        if (varName.equals("$name"))
            if (expr instanceof MethodCall)
                return makeStringL(((MethodCall)expr).getMethodName());
            else if (expr instanceof FieldAccess)
                return makeStringL(((FieldAccess)expr).getFieldName());
            else if (expr instanceof NewExpr)
                return makeStringL(((NewExpr)expr).getClassName());
            else
                throw new CannotCompileException("$name is not available here: "
                                                 + expr.where().toString());

        return null;
    }

    protected static String makeStringL(String str) {
        return "\"" + str + "\"";
    }

    /**
     * Returns an ExprEditor for recursively applying the next advice to
     * $proceed() in the code generated by the current advice. 
     */
    static ExprEditor makeEditor(Hook.Iterator iterator, Expr expr) {
        if (expr instanceof MethodCall)
            return new MethodCallEditor(iterator, (MethodCall)expr);
        else if (expr instanceof FieldAccess)
            return new FieldAccessEditor(iterator, (FieldAccess)expr);
        else if (expr instanceof NewExpr)
            return new NewExprEditor(iterator, (NewExpr)expr);
        else
            throw new RuntimeException("bad join point: " + expr);
    }

    /*
     * ExprEditors for recursively applying other advices.
     */

    static class MethodCallEditor extends ExprEditor {
        private Hook.Iterator advices;
        private String className, methodName, descriptor;

        public MethodCallEditor(Hook.Iterator it, MethodCall call) {
            advices = it;
            className = call.getClassName();
            methodName = call.getMethodName();
            descriptor = call.getSignature();
        }

        public void edit(MethodCall call) throws CannotCompileException {
            if (call.getClassName().equals(className)
                && call.getMethodName().equals(methodName)
                && call.getSignature().equals(descriptor))
                insertHook(advices.copy(), call);
        }
    }

    static class NewExprEditor extends ExprEditor {
        private Hook.Iterator advices;
        private String className, methodName, descriptor;

        public NewExprEditor(Hook.Iterator it, NewExpr call) {
            advices = it;
            className = call.getClassName();
            descriptor = call.getSignature();
        }

        public void edit(NewExpr call) throws CannotCompileException {
            if (call.getClassName().equals(className)
                && call.getSignature().equals(descriptor))
                insertHook(advices.copy(), call);
        }
    }

    static class FieldAccessEditor extends ExprEditor {
        private Hook.Iterator advices;
        private String className, fieldName, descriptor;

        public FieldAccessEditor(Hook.Iterator it, FieldAccess access) {
            advices = it;
            className = access.getClassName();
            fieldName = access.getFieldName();
            descriptor = access.getSignature();
        }

        public void edit(FieldAccess access) throws CannotCompileException {
            if (access.getClassName().equals(className)
                && access.getFieldName().equals(fieldName)
                && access.getSignature().equals(descriptor))
                insertHook(advices.copy(), access);
        }
    }
}
