/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.util.Iterator;
import java.util.TreeMap;

import javassist.CtClass;
import javassist.CtField;
import javassist.ClassPool;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.gluonj.*;

public class GluonMaker {
    /**
     * Reads a glue class to create a Gluon object.
     */
    public Gluon make(String className, ClassPool cp, ClassLoader cloader, Gluon root)
        throws NotFoundException, WeaveException
    {
        Gluon glue = new Gluon(className, cp);
        if (root == null)
            root = glue;

        CtClass clazz = getCtClass(cp, className);
        readOtherGlues(root, clazz, cloader);
        if (Gluon.stackTrace)
           Weaver.logging("Glue: " + className);

        readRefine(glue, clazz);
        readPointcut(glue, clazz, cloader);
        return glue;
    }

    public CtClass getCtClass(ClassPool cp, String name)
        throws NotFoundException
    {
        while (true) {
            try {
                return cp.get(name);
            }
            catch (NotFoundException e) {
                int i = name.lastIndexOf('.');
                if (i < 0)
                    throw e;

                char[] str = name.toCharArray();
                str[i] = '$';
                name = new String(str);
            }
        }
    }

    protected void readOtherGlues(Gluon root, CtClass clazz, ClassLoader cloader)
        throws NotFoundException, WeaveException
    {
        CtClass superClass = clazz.getSuperclass();
        if (superClass != null && !root.isIncluded(superClass.getName())
            && isGlue(superClass))
            root.include(make(superClass.getName(), clazz.getClassPool(), cloader, root));

        TreeMap children = new TreeMap();
        CtField[] fields = clazz.getDeclaredFields();
        for (int i = 0; i < fields.length; i++) {
            CtField f = fields[i];
            Object[] annoList = f.getAvailableAnnotations();
            for (int j = 0; j < annoList.length; j++) {
                Object anno = annoList[j];
                if (anno instanceof Include)
                    children.put(f.getName(), f.getType().getName());
            }
        }

        Iterator it = children.values().iterator();
        while (it.hasNext()) {
            String glueName = (String)it.next();
            if (!root.isIncluded(glueName))
                root.include(make(glueName, clazz.getClassPool(), cloader, root));
        }
    }

    protected void readRefine(Gluon glue, CtClass clazz)
        throws NotFoundException, WeaveException
    {
        CtClass[] nested = clazz.getNestedClasses();
        for (int i = 0; i < nested.length; i++) {
            CtClass cc = nested[i];
            Object[] anno = cc.getAvailableAnnotations();
            boolean notRefiner = true;
            for (int j = 0; j < anno.length; j++)
                if (anno[j] instanceof Refine) {
                    if (!Modifier.isStatic(cc.getModifiers()))
                        throw new WeaveException(
                                "a refine class must be static.");

                    // Refine ref = (Refine)anno[j];
                    glue.refine(cc, anno);
                    notRefiner = false;
                    break;
                }

            if (notRefiner)
                glue.addHelper(cc.getName());
        }
    }

    public static boolean isRefine(CtClass clazz) {
        Object[] anno = clazz.getAvailableAnnotations();
        for (int j = 0; j < anno.length; j++)
            if (anno[j] instanceof Refine)
                return true;

        return false;
    }

    public static boolean isGlue(CtClass clazz) {
        Object[] anno = clazz.getAvailableAnnotations();
        for (int j = 0; j < anno.length; j++)
            if (anno[j] instanceof Glue)
                return true;

        return false;
    }

    /**
     * Reads pointcut-advice pairs and included glues.
     */
    protected void readPointcut(Gluon glue, CtClass clazz, ClassLoader cloader)
        throws WeaveException, NotFoundException
    {
        Object obj = null;
        try {
            Class gclass;
            try {
                String name = clazz.getName();
                if (cloader == null)
                    gclass = Class.forName(name);
                else
                    gclass = cloader.loadClass(name);
            }
            catch (ClassNotFoundException e) {
                gclass = clazz.toClass();
                clazz.defrost();
            }

            Constructor cons = gclass.getDeclaredConstructor(new Class[0]);
            cons.setAccessible(true);
            obj = cons.newInstance(new Object[0]);
        }
        catch (Exception e) {
            throw new WeaveException("cannot instantiate " + clazz.getName(), e);
        }

        CtField[] fields = clazz.getDeclaredFields();
        for (int i = 0; i < fields.length; i++)
            addOneAdvice(glue, clazz, obj, fields[i]);
    }

    protected void addOneAdvice(Gluon glue, CtClass clazz, Object obj, CtField f)
        throws WeaveException, NotFoundException
    {
        Object[] annoList = f.getAvailableAnnotations();
        for (int j = 0; j < annoList.length; j++) {
            Object anno = annoList[j];
            addOneAdvice(glue, clazz, obj, f, anno);
        }
    }

    /*
     * This may return null.
     * It returns an added advice just for extensibility.
     */
    protected Advice addOneAdvice(Gluon glue, CtClass clazz, Object obj,
                                  CtField f, Object anno)
        throws WeaveException, NotFoundException
    {
        String fieldName = f.getName();
        if (anno instanceof Before)
            return glue.addAdvice(Advice.BEFORE, getPointcut(obj, fieldName),
                                 ((Before)anno).value());
        else if (anno instanceof After)
            return glue.addAdvice(Advice.AFTER, getPointcut(obj, fieldName),
                                 ((After)anno).value());
        else if (anno instanceof Around)
            return glue.addAdvice(Advice.AROUND, getPointcut(obj, fieldName),
                                 ((Around)anno).value());
        else if (anno instanceof Include)
            /* ignore */;
        else if (anno instanceof Refine)
            glue.refine(clazz.getClassPool().get(f.getType().getName()), null);
        else
            return unknownAdvice(glue, getPointcut(obj, fieldName), anno);

        return null;
    }

    protected Advice unknownAdvice(Gluon g, Pointcut pc, Object annotation)
        throws WeaveException
    {
        throw new WeaveException("unknown advice");
    }

    protected Pointcut getPointcut(Object obj, String name) throws WeaveException {
        try {
            Field f = obj.getClass().getDeclaredField(name);
            f.setAccessible(true);
            try {
                return (Pointcut)f.get(obj);
            }
            catch (ClassCastException e) {
                throw new WeaveException("not Pointcut type" + f);
            }
        }
        catch (Exception e) {
            throw new WeaveException(e);
        }
    }
}
