/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

/**
 * Load-time weaver.
 */
public class HotSwapper {
    /**
     * Main method.
     *
     * <p>The argument is the name of a glue class.  If the weaver runs in
     * the debug mode, the argument is "&lt;<i>glue name</i>&gt;,debug".
     */
    public static void premain(String args, Instrumentation inst)
        throws WeaveException
    {
        args = parseArgs(args);
        if (args == null || args.length() < 1) {
            fatalError("no glue specified");    // never returns.
            // throw new WeaveException("no glue specified");
        }

        inst.addTransformer(new Transformer(args));
    }

    private static String parseArgs(String args) {
        if (args == null)
            return null;

        int end = args.indexOf(',');
        if (end >= 0) {
            String orig = args;
            args = args.substring(0, end);
            Gluon.stackTrace
                = orig.substring(end).equals(",debug");
        }

        return args.replace('/', '.');
    }

    /**
     * An event handler invoked when a new class file is being loaded.
     */
    public static class Transformer implements ClassFileTransformer {
        private String glueName;
        private Weaver weaver;
        private boolean stop;

        public Transformer(String glue) {
            glueName = glue;
            weaver = null;
            stop = false;
        }

        public byte[] transform(ClassLoader loader, String className, Class classBeingRedefined,
                                ProtectionDomain domain, byte[] classfile)
            throws IllegalClassFormatException
        {
            if (stop)
                return null;

            if (weaver == null) {
                /* The system class loader for java.* and javax.*
                 * may not be able to find a glue class.  Hence,
                 * if loader is the system class loader, then
                 * this method returns without transformation.
                 */
                if (Weaver.isNonTransformable(className.replace('/', '.')))
                    return null;

                try {
                    weaver = new Weaver(glueName, loader);
                }
                catch (Throwable t) {
                    stop = true;
                    showError("cannot read a glue: " + glueName, t);
                    return null;
                }
            }

            try {
                return weaver.transform(className, classfile);
            }
            catch (Throwable t) {
                String msg = "cannot transform a class: " + className.replace('/', '.');
                showError(msg, t);
                return null;
            }
        }

        private void showError(String msg, Throwable e) {
            System.err.println("GluonJ Error: " + msg);
            System.err.println("  by " + e);
            if (Gluon.stackTrace)
                e.printStackTrace(System.err);
        }
    }

    private static void fatalError(String msg) {
        System.err.println("GluonJ Fatal Error: " + msg);
        System.err.println("** exit **");
        System.exit(1);
    }
}
