/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CannotCompileException;
import javassist.NotFoundException;
import javassist.Loader;
import javassist.gluonj.Glue;
import javassist.gluonj.Gluon;
import javassist.gluonj.Refine;
import javassist.gluonj.WeaveException;

/**
 * Compile-time weaver.
 *
 * @see javassist.gluonj.ant.taksdefs.Weave
 */
public class OfflineWeaver {
    private String directory;
    private ArrayList notSavedClasses;
    private Weaver2 weaver;

    /**
     * Constructs an offline weaver.
     * @throws WeaveException 
     */
    public OfflineWeaver(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        weaver = makeWeaver(glueName, classpath);
        directory = ".";
        notSavedClasses = new ArrayList();
    }

    static class Weaver2 extends Weaver {
        Weaver2(String glueName, ClassPool cp, ClassLoader cl)
            throws NotFoundException, WeaveException
        {
            super(glueName, cp, cl);
        }

        protected void runInspectors(CtClass clazz) throws WeaveException {}

        protected void runInspectors0(CtClass clazz) throws WeaveException {
            super.runInspectors(clazz);
        }

        ClassPool getSrcPool() { return srcPool; } 
    }

    private static Weaver2 makeWeaver(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        ClassPool cp = new ClassPool(true);
        ClassPool cp2 = new ClassPool(true);
        if (classpath != null)
            for (int i = 0; i < classpath.length; i++) {
                String path = dirName(classpath[i]);
                cp.appendClassPath(path);
                cp2.appendClassPath(path);
            }

        Loader cloader = new Loader(Gluon.class.getClassLoader(), cp);
        cloader.delegateLoadingOf("javassist.");
        return new Weaver2(glueName, cp2, cloader);
    }

    private static String dirName(String dir) {
        String path = dir.replace(File.separatorChar, '/');
        int len = path.length();
        if (path.charAt(len - 1) == '/')
            path = path.substring(0, len - 1);

        return path;
    }

    /**
     * Sets the output directory where class files are written.
     */
    public void setDirectory(String dir) {
        directory = dirName(dir);
    }

    /**
     * Transforms class files.  It may print an error message.
     */
    public void transformFiles(File baseDir, String[] fileNames)
        throws WeaveException, NotFoundException
    {
        int len = fileNames.length;
        notSavedClasses.ensureCapacity(len);
        ArrayList classes = new ArrayList(len);     // ArrayList<CtClass>
        ClassPool cp = weaver.getSrcPool();
        for (int i = 0; i < len; i++) {
            String fname = fileNames[i];
            if (fname.endsWith(".class"))
                try {
                    readFile(cp, baseDir, fname, classes);
                }
                catch (WeaveException e) {
                    showTransformError(fname, e);
                    throw e;
                }
        }

        Iterator it = classes.iterator();
        while (it.hasNext()) {
            CtClass c = (CtClass)it.next();
            try {
                CtClass cc = weaver.transform(c.getName());
                cc.writeFile(directory);
            }
            catch (WeaveException e) {
                showTransformError(c.getName(), e);
                throw e;
            }
            catch (IOException e) {
                printError(e.getMessage());
                throw new WeaveException(e); 
            }                
            catch (CannotCompileException e) {
                printError(e.getMessage());
                throw new WeaveException(e); 
            }
        }
    }

    private void readFile(ClassPool cp, File baseDir, String name,
                          ArrayList classes)
        throws WeaveException
    {
        try {
            InputStream is
                = new BufferedInputStream(new FileInputStream(new File(baseDir, name)));
            CtClass clazz = cp.makeClass(is);
            is.close();
            if (notGlue(clazz)) {
                weaver.runInspectors0(clazz);
                classes.add(clazz);
            }
        }
        catch (IOException e) {
            throw new WeaveException(e);
        }
    }

    private boolean notGlue(CtClass clazz) {
        boolean result = true;
        Object[] anno = clazz.getAvailableAnnotations();
        for (int j = 0; j < anno.length; j++)
            if (anno[j] instanceof Glue || anno[j] instanceof Refine)
                result = false;

        return result;
    }

    public static void main(String[] args) throws Exception {
        run0(OfflineWeaver.class.getName(), args);
    }

    public static void run(String mainClass, String[] args) {
        try {
            run0(mainClass, args);
        }
        catch (Exception e) {
            if (Gluon.stackTrace)
                e.printStackTrace(System.err);
        }
    }

    private static void run0(String mainClass, String[] args) throws Exception {
        if (args.length < 2) {
            showUsage(mainClass);
            return;
        }

        String directory = ".";
        int index = 0;
        if (args[0].equals("-d")) {
            if (args.length < 4) {
                showUsage(mainClass);
                return;
            }

            directory = args[1];
            index = 2;
        }

        String[] srcDirs = { "." };
        OfflineWeaver w = make(args[index], srcDirs);
        w.setDirectory(directory);
        String[] fileNames = new String[args.length - index - 1];
        for (int i = index + 1; i < args.length; i++)
            fileNames[i - index - 1] = args[i];

        w.transformFiles(null, fileNames);

    }

    public static OfflineWeaver make(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        try {
            return new OfflineWeaver(glueName, classpath);
        }
        catch (NotFoundException e) {
            printError("cannot find a glue " + glueName);
            throw e;
        }
        catch (WeaveException e) {
            printError("cannot read a glue " + glueName);
            throw e;
        }
    }

    private static void showUsage(String mainClass) {
        System.err.println("Usage: java "
                + mainClass + " [ -d <output direcotry> ]"
                + " <glue class name> <java class file> ...");
    }

    public static void showTransformError(String className, WeaveException e) {
        String msg = "cannot transform a class: " + className.replace('/', '.');
        printError(msg);
        System.err.println("  by " + e);
    }

    private static void printError(String msg) {
        System.err.println("GluonJ Error: " + msg);
    }
}
