/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.util.HashMap;
import java.util.ArrayList;
import javassist.*;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ConstantAttribute;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.AttributeInfo;
import javassist.bytecode.annotation.Annotation;
import javassist.convert.Transformer;
import javassist.gluonj.Super;
import javassist.gluonj.WeaveException;

/**
 * Refine class processor.
 */
public class RefineWeaver {
    /**
     * Returns the target class of the given refine class.
     */
    public CtClass getTarget(CtClass refiner) throws WeaveException {
        try {
            return getTarget0(refiner);
        }
        catch (NotFoundException nfe) {
            throw new WeaveException(nfe);
        }
    }

    protected CtClass getTarget0(CtClass refiner) throws WeaveException, NotFoundException {
        boolean isIntf = refiner.isInterface();
        if (isIntf) {
            CtClass[] interfaces = refiner.getInterfaces();
            if (interfaces.length > 0)
                return interfaces[0];
            else
                throw new WeaveException(refiner.getName()
                            + " must extend another interface");
        }
        else {
            CtClass target = refiner.getSuperclass();
            if (target.getName().equals("java.lang.Object"))
                throw new WeaveException(refiner.getName()
                            + " must not directly extend java.lang.Object");

            return target;
        }
    }

    /**
     * Appends the interfaces and methods declared in the
     * class to its super class.  <code>refiner</code> will be
     * modified; it should be discarded after this method returns.
     *
     * @param refiner      the class
     * @return the modified super class.
     */
    public void weave(CtClass target, Refiner refiner, ClassMap allRefinerNames)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtClass refinerClass = refiner.clazz;
        if (refiner.targetClass != null && !refiner.privileged)
            throw new WeaveException("@Refine(" + refiner.targetClass
                                     + ") needs @Privileged: " + refinerClass.getName());

        boolean isIntf = refinerClass.isInterface();
        copyFields(target, refiner);
        CtClass[] interfaces = refinerClass.getInterfaces();
        int first = isIntf ? 1 : 0;
        for (int i = first; i < interfaces.length; i++)
            if (!target.subtypeOf(interfaces[i]))
                target.addInterface(interfaces[i]);

        if (!isIntf)
            copyConstructors(target, refiner, allRefinerNames);

        copyClassInitializer(target, refiner, allRefinerNames);
        copyMethods(target, refiner, isIntf, allRefinerNames);
        refiner.revert();
    }

    protected void copyFields(CtClass target, Refiner refiner)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtClass refinerClass = refiner.clazz;
        CtField[] newFields = refinerClass.getDeclaredFields();
        for (int i = 0; i < newFields.length; i++) {
            CtField f = newFields[i];
            CtField origF = null;
            try {
                origF = target.getField(f.getName());
            }
            catch (NotFoundException e) {
                if (isSuperAlias(f) != null)
                    throw new WeaveException("non-overriding field with @Super: " + f);

                CtField newF = new CtField(f, target);
                target.addField(newF);
                continue;     // not override any superclass field
            }

            /* Defining a field with the same name as the super's field
             * is not allowed for the time being.
             * This is different from the Java's semantics.
             */
            if (!refiner.privileged && !origF.visibleFrom(refinerClass))
                throw new WeaveException(refinerClass.getName() + " cannot access a field", origF);

            if (!origF.getSignature().equals(f.getSignature()))
                throw new WeaveException("incompatible field type", f);

            if (Modifier.isFinal(origF.getModifiers()))
                throw new WeaveException("cannot update a final field", origF);

            // copy attributes.
            /* If a field overrides its super's field, then the annotations of the former
               field are appended to the super's field.
             */
            FieldInfo info = f.getFieldInfo2();
            FieldInfo origInfo = origF.getFieldInfo2();
            AttributeInfo ai = info.getAttribute(ConstantAttribute.tag);
            if (ai != null)
                origInfo.addAttribute(ai);

            appendAnnotations(true, info, origInfo);
            appendAnnotations(false, info, origInfo);
        }
    }

    private static void appendAnnotations(boolean visible,
                                          FieldInfo src, FieldInfo dest)
    {
        String tag = visible ? AnnotationsAttribute.invisibleTag
                             : AnnotationsAttribute.visibleTag; 
        AnnotationsAttribute srcAttr
            = (AnnotationsAttribute)src.getAttribute(tag);
        if (srcAttr == null)
            return;

        AnnotationsAttribute destAttr
            = (AnnotationsAttribute)dest.getAttribute(tag);
        if (destAttr == null) {
            dest.addAttribute(srcAttr);
            return;
        }

        Annotation[] anno = srcAttr.getAnnotations();
        for (int i = 0; i < anno.length; i++)
            destAttr.addAnnotation(anno[i]);
    }

    /* The refiner must have only a constructor taking no argument.
     * The super() call in that constructor body is eliminated.
     */
    protected static void copyConstructors(CtClass target, Refiner refiner, ClassMap allRefiners)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtClass refinerClass = refiner.clazz;
        CtConstructor[] cons = refinerClass.getDeclaredConstructors();
        if (cons.length != 1)
            throw new WeaveException("more than one constructors in "
                                     + refinerClass.getName());

        CtConstructor cc = cons[0];
        if (cc.isEmpty())       // defualt constructor
            return;

        if (!cc.getSignature().equals("()V"))
            throw new WeaveException("constructor with parameters in " + refinerClass.getName());

        String methodName = target.makeUniqueName("_init");
        CtMethod mth = cc.toMethod(methodName, target, allRefiners);
        mth.setModifiers(Modifier.PRIVATE);
        target.addMethod(mth);

        String flagName = target.makeUniqueName("_init_flag");
        CtField flag = new CtField(CtClass.booleanType, flagName, target);
        target.addField(flag);
        cons = target.getDeclaredConstructors();
        String expr = "if(!" + flagName + "){" + flagName + "=true;" + methodName + "();}";
        for (int i = 0; i < cons.length; i++) {
            CtConstructor c = cons[i];
            c.insertAfter(expr);
        }
    }

    protected static void copyClassInitializer(CtClass target, Refiner refine, ClassMap allRefiners)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtConstructor cons = refine.clazz.getClassInitializer();
        if (cons == null || cons.isEmpty())
            return;

        CtConstructor dest = target.getClassInitializer();
        if (dest == null) {
            dest = new CtConstructor(cons, target, allRefiners);
            target.addConstructor(dest);
        }
        else {
            String methodName = target.makeUniqueName("_clinit");
            CtMethod mth = cons.toMethod(methodName, target, allRefiners);
            mth.setModifiers(Modifier.PRIVATE | Modifier.STATIC);
            target.addMethod(mth);
            dest.insertAfter(methodName + "();");
        }
    }

    public static class MethodDef {
        public MethodDef(CtMethod m) {
            target = m;
            overrider = null;
        }

        public CtMethod target;

        // non-null if another @Refine has already redefined the target.
        public CtMethod overrider;
    }

    /**
     * @param allRefiners           this is modified by copyMethods().
     */
    protected void copyMethods(CtClass target, Refiner refiner, boolean isIntf,
                               ClassMap allRefiners)
        throws WeaveException, NotFoundException, CannotCompileException
    {
        /* Overriding an invisible method is not allowed unless privileged is on.
         * For example,
         *
         * public class foo.A {
         *   int foo() { return 0; }  // package only
         * }
         * public class bar.B extends foo.A {
         *   public int foo() { return 1; }
         * }
         *
         * foo() in B does not override foo() in A.  It would be OK
         * to make the two foo()s coexist in A after weaving but
         * that design would be also error prone. 
         */
        HashMap hash = new HashMap();
        CtMethod[] origMethods = target.getDeclaredMethods();
        for (int i = 0; i < origMethods.length; i++) {
            CtMethod m = origMethods[i];
            hash.put(m, new MethodDef(m));
        }

        readParentMethods(refiner.parent, hash);

        CtClass refinerClass = refiner.clazz;
        CtMethod[] newMethods = refinerClass.getDeclaredMethods();
        ArrayList superOrigPairs = null;
        superOrigPairs = findSuperMethods(refiner, target, newMethods);
        CtMethod[] addedMethods = new CtMethod[newMethods.length];
        int numAdded = 0;
        CodeConverter refinerConv = new CodeConverter2();

        numAdded = copyMethodLoop(target, isIntf, refiner,
                                  hash, newMethods, addedMethods, numAdded, refinerConv);

        if (superOrigPairs != null)
            convSuperToOrig(refinerConv, superOrigPairs);

        allRefiners.fix(target);
        for (int i = 0; i < numAdded; i++) {
            CtMethod m = addedMethods[i];
            m.instrument(refinerConv);          // target should be detached.
            CtMethod newM = CtNewMethod.copy(m, target, allRefiners);
            target.addMethod(newM);
        }

        beAbstractIfYes(target);
    }

    private void readParentMethods(Refiner r, HashMap hash)
        throws WeaveException, NotFoundException, CannotCompileException
    {
        if (r != null) {
            readParentMethods(r.parent, hash);
            CtMethod[] mths = r.clazz.getDeclaredMethods();
            for (int i = 0; i < mths.length; i++) {
                CtMethod m = mths[i];
                MethodDef def = (MethodDef)hash.get(m);
                if (def != null)
                    def.overrider = m;
            }
        }
    }

    protected void beAbstractIfYes(CtClass clazz) {
        int mod = clazz.getModifiers();
        if (Modifier.isInterface(mod))
            return;

        CtMethod[] m = clazz.getDeclaredMethods();
        int n = m.length;
        for (int i = 0; i < n; i++)
            if (Modifier.isAbstract(m[i].getModifiers())) {
                clazz.setModifiers(clazz.getModifiers() | Modifier.ABSTRACT);
                return;
            }

        clazz.setModifiers(Modifier.clear(mod, Modifier.ABSTRACT));
    }

    protected int copyMethodLoop(CtClass target, boolean isIntf, Refiner refiner,
                                 HashMap hash, CtMethod[] newMethods,
                                 CtMethod[] addedMethods, int numAdded, CodeConverter refinerConv)
        throws WeaveException, CannotCompileException
    {
        CtClass refinerClass = refiner.clazz;
        boolean privileged = refiner.privileged;
        for (int i = 0; i < newMethods.length; i++) {
            CtMethod m = newMethods[i];
            if (isSuperAlias(m) != null)
                continue;

            MethodDef found = (MethodDef)hash.get(m);
            if (found != null) {
                CtMethod mth = found.target;
            	if (!privileged && !mth.visibleFrom(refinerClass)) 
            		throw new WeaveException("cannot override an invisible method: "
                            + m.getName() + " in " + refinerClass.getName());

                if (!isIntf) {
                    if (Modifier.isAbstract(mth.getModifiers()))
                        try {
                            target.removeMethod(mth);
                        }
                        catch (NotFoundException e) {}
                    else
                        renameOrigMethod(target, found, refinerConv);

                    addedMethods[numAdded++] = m;
                }
            }
            else
                addedMethods[numAdded++] = m;
        }

        return numAdded;
    }

    private String isSuperAlias(CtMember m) {
        Object[] anno = m.getAvailableAnnotations();
        for (int i = 0; i < anno.length; i++)
            if (anno[i] instanceof Super)
                return ((Super)anno[i]).value();

        return null;
    }

    private ArrayList findSuperMethods(Refiner refiner, CtClass target,
                                       CtMethod[] newMethods)
        throws WeaveException
    {
        ArrayList list = new ArrayList();
        findSuperMethods0(list, refiner, target, newMethods);
        while (refiner.parent != null) {
            refiner = refiner.parent;
            findSuperMethods0(list, refiner, target,
                              refiner.clazz.getDeclaredMethods());
        }

        return list;
    }

    private void findSuperMethods0(ArrayList superOrigPairs, Refiner refiner, CtClass target,
                                   CtMethod[] newMethods)
        throws WeaveException
    {
        boolean privileged = refiner.privileged;
        for (int j = 0; j < newMethods.length; j++) {
            CtMethod m = newMethods[j];
            String name = m.getName();
            String origName = isSuperAlias(m);
            if (origName != null) {
                if (origName.equals(""))
                    origName = name;

                try {
                    CtMethod origM = target.getMethod(origName, m.getSignature());
                    if (origM.getDeclaringClass() == target) {
                        if (!privileged && !origM.visibleFrom(refiner.clazz))
                            throwBadSuper(name, refiner.clazz);
                    }
                    else if (!origM.visibleFrom(target))
                        throwBadSuper(name, refiner.clazz);

                    superOrigPairs.add(m);
                    superOrigPairs.add(origM);
                }
                catch (NotFoundException e) {
                    String msg = "no such @Super method: " + name + " in " + refiner.clazz.getName();
                    throw new WeaveException(msg);
                }
            }
        }
    }

    private void throwBadSuper(String name, CtClass refiner) throws WeaveException {
        String msg = "invisible @Super method: " + name + " in " + refiner.getName();
        throw new WeaveException(msg);
    }

    private void convSuperToOrig(CodeConverter conv, ArrayList superOrigPairs)
        throws CannotCompileException
    {
        int n = superOrigPairs.size();
        int i = 0;
        while (i < n) {
            CtMethod superM = (CtMethod)superOrigPairs.get(i++);
            CtMethod origM = (CtMethod)superOrigPairs.get(i++);
            conv.redirectMethodCall(superM, origM);
        }
    }

    protected static final String origPrefix = "_orig_";

    private void renameOrigMethod(CtClass orig, MethodDef origM, CodeConverter conv)
        throws CannotCompileException
    {
        CtMethod mth = origM.target;
        String origName = mth.getName();
        String methodName = origPrefix + origName;
        try {
            orig.getMethod(methodName, mth.getSignature());
            // unless an exception is thrown, methodName is already used.
            methodName = orig.makeUniqueName(methodName);
        }
        catch (NotFoundException e) {}
        mth.setName(methodName);
        int mod = mth.getModifiers();
        mth.setModifiers(Modifier.setPrivate(mod));
        if (origM.overrider == null)
            conv.redirectMethodCall(origName, mth);
        else
            conv.redirectMethodCall(origM.overrider, mth);
    }

    public static class CodeConverter2 extends CodeConverter {
        public void redirectMethodCall(CtMethod oldMethod, CtMethod newMethod)
            throws CannotCompileException
        {
            transformers = new TransformSuperCall(transformers,
                                                  oldMethod, newMethod);
        }
    }

    public static class TransformSuperCall extends javassist.convert.TransformCall {
        public TransformSuperCall(Transformer next,
                                  CtMethod origMethod, CtMethod substMethod) {
            super(next, origMethod, substMethod);
            // enforce using invokespecial to call the new method
            newMethodIsPrivate = true;
        }
    }
}
