/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import javassist.*;
import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;

/**
 * A glue processor.  It runs a RefineWeaver and an AdviceWeaver for
 * processing a glue.
 *
 * @see OfflineWeaver, HotSwapper
 * @see RefineWeaver, AdviceWeaver
 */
public class Weaver {
    protected ClassPool srcPool;      // original classes
    protected ClassPool refinedPool;  // refined classes (advice is not woven yet)

    protected RefineWeaver refineWeaver;
    protected AdviceWeaver adviceWeaver;
    protected Gluon glue;
    protected HashMap refineTargets;    // HashMap<String target, LinkedList<Refiner>>
    protected ClassMap refinerNames;

    public static void logging(String msg) {
        System.err.print("(GluonJ) ");
        System.err.println(msg);
    }

    /**
     * Constructs a weaver.
     *
     * @param glueName      the name of the root glue.
     * @param loader        the class loader that is used for reading
     *                      original class files.  It is also used
     *                      for loading the glue classes.
     */
    public Weaver(String glueName, ClassLoader loader)
        throws WeaveException
    {
        this(glueName, makeClassPool(loader), loader);
    }

    private static ClassPool makeClassPool(ClassLoader loader) {
        ClassPool cp = new ClassPool(true);
        if (loader != null)
            cp.insertClassPath(new LoaderClassPath(loader));

        return cp;
    }

    /**
     * Constructs a weaver.  cp is used for reading original class files.
     * Appropriate <code>javassist.ClassPath</code> objects must be registered
     * to this class pool.
     *
     * @param glueName  the name of the root glue.
     * @param cp        the pool of the classes created from original class files.
     * @param loader    the class loader for glues.  If this is null,
     *                  glues are loaded by Class.forName().
     * @throws WeaveException 
     */
    public Weaver(String glueName, ClassPool cp, ClassLoader loader)
        throws WeaveException
    {
        srcPool = cp;
        refinedPool = makePool();
        Gluon g = Gluon.make(glueName, cp, loader);

        initWeavers();
        glue = g;
        refineTargets = new HashMap();
        refinerNames = loadRefines(g);
    }

    /**
     * Returns the source ClassPool.
     */
    public ClassPool getSource() { return srcPool; }

    /**
     * Returns the ClassPool in which classes are already refined.
     */
    public ClassPool getRefined() { return refinedPool; }

    protected void initWeavers() {
        refineWeaver = new RefineWeaver();
        adviceWeaver = new AdviceWeaver();
    }

    /*
     * This method reads the definitions of @Refine classes and
     * computes their target classes etc.  A Gluon object does
     * not care about them.
     */
    private ClassMap loadRefines(Gluon glue) throws WeaveException {
        ClassMap map = new ClassMap();
        HashMap refiners = new HashMap();
        Iterator iterator = glue.getRefiners();
        while (iterator.hasNext()) {
            Refiner r = (Refiner)iterator.next();
            CtClass target = getRefineTarget(r);
            if (GluonMaker.isRefine(target)) {
                Refiner parent = (Refiner)refiners.get(target);
                if (parent == null)
                    throw new WeaveException(target.getName()
                            + " must be declared before " + r.clazz.getName());
                else {
                    r.parent = parent;
                    target = parent.realTarget;
                }
            }

            String targetName = target.getName();
            addRefine(targetName, r);
            String thisName = r.clazz.getName();
            map.put(thisName, targetName);
            r.realTarget = target;
            refiners.put(r.clazz, r);
        }

        return map.isEmpty() ? null : map;
    }

    private CtClass getRefineTarget(Refiner ref) throws WeaveException {
        if (ref.targetClass == null)
            return refineWeaver.getTarget(ref.clazz);
        else
            try {
                return ref.clazz.getClassPool().get(ref.targetClass);
            }
            catch (NotFoundException e) {
                throw new WeaveException(e.getMessage());
            }
    }

    private void addRefine(String target, Refiner refine) {
        LinkedList entry = findRefiners(target);
        if (entry == null)
            entry = new LinkedList();

        entry.addLast(refine);
        refineTargets.put(target, entry);
    }

    private LinkedList findRefiners(String targetName) {
        Object refines = refineTargets.get(targetName);
        if (refines == null) {
            String name = targetName.replace('$', '.');
            if (name != targetName) {
                refines = refineTargets.get(name);
                targetName = name;
            }
        }

        // if (refines != null)
        //    refineTargets.remove(targetName);

        return (LinkedList)refines;
    }

    /**
     * Returns a glue.
     */
    public Gluon getGlue() { return glue; }

    /**
     * Transforms the class with a given name.
     * The class file is read from the class pool.
     * If <code>isNonTransformable(className)</code> is true,
     * this method does not perform transformation.
     */
    public CtClass transform(String className)
        throws WeaveException, NotFoundException
    {
        try {
            CtClass clazz;
            if (!isNonTransformable(className))
                clazz = weaveAdvices(className);
            else
                clazz = srcPool.get(className);

            return clazz;
        }
        catch (CannotCompileException cce) {
            throw new WeaveException("cannot transform a class file: "
                                     + toClassName(className), cce);
        }
    }

    /**
     * Transforms a given class file.  It returns null if no transformation
     * is needed.
     */
    public byte[] transform(String className, byte[] classFile)
        throws WeaveException
    {
        className = toClassName(className);
        if (isNonTransformable(className))
            return null;

        try {
            /* If multiple class loaders use the same Weaver object
             * for reading a class file, the same class file might
             * be processed more than once.
             */
            CtClass original
                = srcPool.makeClass(new ByteArrayInputStream(classFile), false);
            className = original.getName();
            CtClass clazz = weaveAdvices(className);
            return clazz.toBytecode();
        }
        catch (IOException ie) {
            throw new WeaveException("cannot read a class file: "
                                     + className, ie);
        }
        catch (CannotCompileException cce) {
            throw new WeaveException("cannot transform a class file: "
                                     + className, cce);
        }
    }

    /**
     * Returns true if the given class name represents a system class.
     * Note that JVM does not accept transformed system classes.
     */
    protected static boolean isNonTransformable(String className) {
        return className.startsWith("java.")
            || className.startsWith("javax.")
            || className.startsWith("com.sun.")
            || className.startsWith("sun.")
            || className.startsWith("sunw.");
            // || className.startsWith("javassist.");
    }

    private static String toClassName(String jvmName) {
        return jvmName.replace('/', '.');
    }

    /**
     * Weaves advices into a target class.
     */
    protected CtClass weaveAdvices(String targetName)
        throws WeaveException, CannotCompileException
    {
        CtClass target;
        try {
            // target has been already refined.
            target = refinedPool.get(targetName);
        }
        catch (NotFoundException ne) {
            throw new WeaveException("cannot find: " + targetName);
        }

        try {
            if (!target.isFrozen()) {
                adviceWeaver.weave(target, glue);
                if (glue.isHelper(target.getName())) {
                    target.replaceClassName(refinerNames);
                }
            }

            return target;
        }
        catch (WeaveException e) {
            e.setWhere("While transforming " + targetName);
            throw e;
        }
    }

    protected void refineClass(CtClass target)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        try {
            CtClass sup = target.getSuperclass();
            if (sup != null && !isNonTransformable(sup.getName()) && !sup.isFrozen()) {
                refineClass(sup);
                sup.freeze();
            }
        }
        catch (NotFoundException e) {
            /* getSuperclass() may throw an exception.
             * If it does, we ignore the exception.
             */
            if (Gluon.stackTrace)
                Weaver.logging("[warning] cannot find a class file: " + e.getMessage());
        }

        refineClass0(target);
    }

    protected void refineClass0(CtClass target)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        ClassMap names = refinerNames; 
        runInspectors(target);
        LinkedList refiners = findRefiners(target.getName());
        if (refiners == null)
            return;

        Iterator it = refiners.iterator();
        while (it.hasNext()) {
            Refiner refiner = (Refiner)it.next();
            refineWeaver.weave(target, refiner, names);
        }
    }

    /**
     * Let glues inspect a class definition before weaving.
     */
    protected void runInspectors(CtClass clazz)
        throws WeaveException
    {
        List inspectors = glue.getInspectors();
        if (inspectors == null)
            return;

        Iterator it = inspectors.iterator();
        while (it.hasNext()) {
            Inspector inspector = (Inspector)it.next();
            inspector.inspect(clazz);
        }
    }

    private ClassPool makePool() {
        ClassPool cp = new ClassPool();
        ClassPath path = new ClassPath() {
            public void close() {}

            public URL find(String classname) {
                URL url = srcPool.find(classname);
                if (url == null)
                    try {
                        srcPool.get(classname);
                        url = new URL("file:memory");
                    }
                    catch (NotFoundException e) {
                        return null;
                    } catch (MalformedURLException e) {
                        return null;
                    }

                return url;
            }

            public InputStream openClassfile(String classname) throws NotFoundException {
                CtClass cc = srcPool.get(classname);
                Exception err = null;
                try {
                    if (!isNonTransformable(classname) && !cc.isFrozen())
                        refineClass(cc);

                    byte[] file = cc.toBytecode();
                    return new java.io.ByteArrayInputStream(file);
                }
                catch (IOException ie) { err = ie; }
                catch (CannotCompileException cce) { err = cce; }
                catch (WeaveException we) { err = we; }
                throw new NotFoundException(classname, err);
            }
        };
        cp.appendClassPath(path);
        return cp;
    }
}
