/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006-2007 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.ClassMap;
import javassist.NotFoundException;
import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Not used any more.
 *
 * @deprecated
 */
public class WeaverCore {
    private Gluon glue;
    protected ClassPool pool;
    protected RefineWeaver refineWeaver;
    protected AdviceWeaver adviceWeaver;
    protected HashMap refineTargets;    // HashMap<String target, LinkedList<Refiner>>
    protected ClassMap refinerNames;

    /**
     * Constructs a weaver.
     */
    public WeaverCore(Gluon g) throws WeaveException {
        glue = g;
        pool = g.getPool();
        refineTargets = new HashMap();
        initWeavers();
        refinerNames = loadRefines();
    }

    protected void initWeavers() {
        refineWeaver = new RefineWeaver();
        adviceWeaver = new AdviceWeaver();
    }

    private ClassMap loadRefines() throws WeaveException {
        ClassMap map = new ClassMap();
        Iterator iterator = glue.getRefiners();
        while (iterator.hasNext()) {
            Refiner r = (Refiner)iterator.next();
            CtClass target = refineWeaver.getTarget(r.clazz);
            String targetName = target.getName();
            addRefine(targetName, r);
            map.put(r.clazz.getName(), targetName);
        }

        return map.isEmpty() ? null : map;
    }

    private void addRefine(String target, Refiner refine) {
        LinkedList entry = findRefiners(target, false);
        if (entry == null)
            entry = new LinkedList();

        entry.addLast(refine);
        refineTargets.put(target, entry);
    }

    private LinkedList findRefiners(String targetName, boolean delete) {
        Object refines = refineTargets.get(targetName);
        if (refines == null) {
            String name = targetName.replace('$', '.');
            if (name != targetName) {
                refines = refineTargets.get(name);
                targetName = name;
            }
        }

        if (delete && refines != null)
            refineTargets.remove(targetName);

        return (LinkedList)refines;
    }

    protected ClassPool getClassPool() {
        return glue.getPool();
    }

    protected Gluon getGlue() { return glue; }

    /**
     * Transforms a given class.  Refine classes are removed after
     * they are applied to their target class. 
     *
     * @param targetName        fully-qualified class name.
     */
    public boolean transform(String targetName) throws WeaveException {
        return transform(targetName, true);
    }

    /**
     * @param inspect       if it is true, let glues examine the definition of
     *                      the transformed class.
     */
    public boolean transform(String targetName, boolean inspect)
        throws WeaveException
    {
        try {
            return transform(targetName, findRefiners(targetName, true), inspect);
        }
        catch (WeaveException e) {
            e.setWhere("While transforming " + targetName);
            throw e;
        }
    }

    /**
     * Makes sure that all refine classes have been applied to target classes.
     * Otherwise, an exception is thrown.
     */
    public void done() throws WeaveException {
        if (refineTargets.size() == 0)
            return;

        Iterator it = refineTargets.entrySet().iterator();
        StringBuffer sbuf = new StringBuffer("some refine classes were not used: ");
        boolean first = true;
        while (it.hasNext()) {
            LinkedList list = (LinkedList)(((java.util.Map.Entry)it.next()).getValue());
            Iterator it2 = list.iterator();
            while (it2.hasNext()) {
                if (first)
                    first = false;
                else
                    sbuf.append(", ");

                Refiner r = (Refiner)it2.next();
                sbuf.append(r.clazz.getName());
            }
        }

        throw new WeaveException(sbuf.toString());
    }

    /**
     * Applies given refine classes to a target class
     * and then weaves advices.
     *
     * @param refines       a list of refine classes obtained from
     *                      refineTargets.
     * @return              true if the target class was modified. 
     */
    protected boolean transform(String targetName, LinkedList refines, boolean inspect)
        throws WeaveException
    {
        try {
            CtClass target = pool.get(targetName);
            if (inspect)
                runInspectors(target, glue.getInspectors());

            boolean modify = refineClass(refineWeaver, target, refines);
            boolean modify2 = weaveAdvices(target);
            boolean result = modify || modify2; 
            save(target, result);
            return result;
        }
        catch (IOException ioe) {
            throw new WeaveException(ioe);
        }
        catch (CannotCompileException cce) {
            throw new WeaveException(cce);
        }
        catch (NotFoundException nfe) {
            throw new WeaveException(nfe);
        }
    }

    /**
     * Let glues inspect a class definition before weaving.
     */
    protected void runInspectors(CtClass clazz, List inspectors)
        throws WeaveException
    {
        if (inspectors == null)
            return;

        Iterator it = inspectors.iterator();
        while (it.hasNext()) {
            Inspector inspector = (Inspector)it.next();
            inspector.inspect(clazz);
        }
    }

    /**
     * Applies refine classes to a target class.
     *
     * @param refines   a list of refine classes.
     */
    protected boolean refineClass(RefineWeaver rw, CtClass target,
                                  LinkedList refines)
        throws WeaveException, NotFoundException, CannotCompileException
    {
        if (refines == null)
            return false;

        boolean modified = false;
        Iterator it2 = refines.iterator();
        while (it2.hasNext()) {
            Refiner refiner = (Refiner)it2.next();
            CtClass refinerClass = refiner.clazz;
            rw.weave(target, refiner, refinerNames);
            refinerClass.detach();
            modified = true;
        }

        return modified;
    }

    /**
     * Weaves advices into a target class.
     */
    protected boolean weaveAdvices(CtClass target)
        throws WeaveException, CannotCompileException
    {
        boolean result = adviceWeaver.weave(target, glue);
        if (glue.isHelper(target.getName())) {
            target.replaceClassName(refinerNames);
            result = result ? true : target.isModified();
        }

        return result;
    }

    /**
     * Finalizes the transformation of a given class.
     * A subclass can overwrite this method.
     * The default implementation does nothing.
     *
     * @param modify    the class was modified.
     */
    protected void save(CtClass clazz, boolean modify)
        throws CannotCompileException, IOException
    {
    }
}
